
import asyncio
import nest_asyncio
import aiohttp
import sys
import logging
import contextlib
from time import sleep

from aiohttp.client_exceptions import ServerDisconnectedError, ClientConnectionError
from asyncio.exceptions import TimeoutError
from urllib.error import URLError

from .base_scrape import BaseScrape
from ..utils.header_vars import random_header_vars
from ..utils.errors import HttpResponseStatusError


class AsyncScrape(BaseScrape):
    def __init__(self,
                 post_process_func: callable,
                 post_process_kwargs: dict = {},
                 fetch_error_handler: callable = None,
                 use_proxy: bool = False,
                 proxy: str = None,
                 pac_url: str = None,
                 consecutive_error_limit: int = 100,
                 attempt_limit: int = 5,
                 rest_between_attempts: bool = True,
                 rest_wait: int = 60,
                 randomise_headers: bool = False
                 ):
        """Class for scrapping webpages

        args:
        ----
        post_process_func - callable - for processing html
        post_process_kwargs - dict:{} - kwargs for use in post processing
        fetch_error_handler - callable:None - the function to be called if an
            error is experienced during _fetch. Passes in:
            url, error as arguments
        use_proxy - bool:False - should a proxy be used
        proxy - str:None - what is the address of the proxy ONLY VALID IF
            PROXY IS TRUE
        pac_Url - str:None - the location of the pac information ONLY VALID IF
            PROXY IS TRUE
        consecutive_error_limit - int:100 - the number of times an error can be experienced 
            in a row before the scrape is cancelled and a new round is started
        attempt_limit - int:5 - number of times a url can be attempted before it's abandoned
        rest_between_attempts - bool:True - should the program rest between scrapes
        rest_wait - int:60 - how long should the program rest for ONLY VALID IF
            REST_BETWEEN_SCRAPES IS TRUE
        randomise_headers - bool:False - should the headers be randomised after each request
        """
        # Init super
        super().__init__(
            use_proxy=use_proxy,
            proxy=proxy,
            pac_url=pac_url
        )
        self.post_process = post_process_func
        self.post_process_kwargs = post_process_kwargs
        self.randomise_headers = randomise_headers
        self.headers = random_header_vars(self.header_vars)
        self.fetch_error_handler = fetch_error_handler
        self.shutdown_initiated = False
        # Establish loop and coro
        self.loop = None
        self.coro = None
        self.gathered_tasks = None
        # Define allowed errors
        self.acceptable_errors = (ServerDisconnectedError, ClientConnectionError,
                                  TimeoutError, URLError, HttpResponseStatusError)
        self.consecutive_error_limit = consecutive_error_limit
        self.consecutive_error_count = 0
        # Define criteria for looping multiple attempts
        self.attempt_limit = attempt_limit
        self.rest_between_attempts = rest_between_attempts
        self.rest_wait = rest_wait
        self.tracker = None
        self.cur_err = None

    async def shutdown(self):
        # Mark shutdown as started
        self.shutdown_initiated = True
        logging.info("Shutdown of scrape initialized ...")
        self.gathered_tasks.cancel()

    def _proxy(self):
        # Set policy if using windows
        self._set_policy()
        # Start the pac session
        self._get_pac_session()

    def _set_policy(self):
        if sys.platform.startswith("win") \
                and sys.version_info[0] == 3 \
                and sys.version_info[1] >= 8:
            policy = asyncio.WindowsSelectorEventLoopPolicy()
            asyncio.set_event_loop_policy(policy)

    def _set_event_loop(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self.loop = asyncio.get_event_loop()

    def _get_event_loop(self):
        self._set_event_loop()
        if isinstance(self.loop, asyncio.BaseEventLoop):
            nest_asyncio.apply()
        return self.loop

    async def _fetch(self, session, url):
        """Function to fetch HTML from url

        args:
        ----
        session - aiohttp.ClientSession() object
        url - str - url to be requested

        returns:
        ----
        list
        """
        # Regenerate headers
        if self.randomise_headers:
            self.headers = random_header_vars(self.header_vars)
        # Get the proxy for this url
        proxy = self._get_proxy(url)
        status = None
        # Fetch with aiohttp session
        try:
            async with session.request("get", url, proxy=proxy, headers=self.headers) as resp:
                status = resp.status
                if status == 200:
                    html = await resp.text()
                    func_resp = self.post_process(
                        html=html, resp=resp, **self.post_process_kwargs)
                else:
                    raise HttpResponseStatusError(
                        f"url responded with a http status of {status}")
                # Reset self.acceptable_error_count if all goes fine
                self.consecutive_error_count = 0
                return {"url": url, "func_resp": func_resp, "status": resp.status, "error": None}
        except Exception as e:
            # Set the current error - increment if the same error
            if type(e) == self.cur_err:
                self.consecutive_error_count += 1
            else:
                self.cur_err = type(e)
                self.consecutive_error_count = 1
            # Check if acceptabe error limit has been reached
            # this prevents functions from carrying on after a site has started blocking calls
            if self.consecutive_error_count >= self.consecutive_error_limit \
                    and not self.shutdown_initiated:
                await self.shutdown()
                logging.warning(
                    f"Consecutive error limit reached - {e} - consecutive count at {self.consecutive_error_count}/{self.consecutive_error_limit}")
            # Check for error handler
            if self.fetch_error_handler:
                logging.info(
                    f"Error passed to {self.fetch_error_handler.__name__}")
                # Run the error handler
                self.fetch_error_handler(url, e)
            # Check if acceptable error
            if type(e) in self.acceptable_errors:
                logging.warning(
                    f"Acceptable error in request or post processing {url} - {e}")
            # Raise error
            else:
                logging.error(
                    f"Unhandled error in request or post processing {url} - {e}")
                if f"{e}" == "":
                    raise e
            return {"url": url, "func_resp": None, "status": status, "error": e}

    async def _fetch_async(self,
                           session, url
                           ):
        """Function for getting routes from one locationt to another by different 
        modes of transport.

        args:
        ----
        session - aiohttp.ClientSession() object
        url - str - url to be requested

        returns:
        ----
        float - time (hours)
        """
        # Establish return object
        rtrn = {
            "url": url,
            "func_resp": None,
            "status": None
        }
        # Fetch
        rtrn = await self._fetch(session, url)
        # Increment the pages scraped
        self.increment_pages_scraped()
        return rtrn

    async def _fetch_all_async(self, urls):
        """"Async function for finding the latitude and 
        longitude of a list of locations

        args:
        ----
        urls - list - the pages to be scraped

        returns:
        ----
        gathered asyncio tasks
        """
        tasks = []
        async with aiohttp.ClientSession() as session:
            for url in urls:
                tasks.append(
                    self._fetch_async(
                        session,
                        url
                    )
                )
            self.gathered_tasks = asyncio.gather(*tasks)
            with contextlib.suppress(asyncio.CancelledError):
                _ = await self.gathered_tasks
            # Get results individually to deal with premature cancellation
            resps = [
                r.result() for r in self.gathered_tasks._children
                if not r.cancelled()
            ]
            return resps

    # run from terminal
    def scrape_all(self, urls: list = []):
        """"Function asynchronously scraping html from urls and passing 
        them through the post processing function

        args:
        ----
        urls - list - the pages to be scraped

        returns:
        ----
        list of dicts
            EG [{
                "url":"http://google.com",
                "func_resp":response from post process function,
                "status":200
            }]
        """
        self.start_job()
        if self.use_proxy:
            self._proxy()
        # Establish urls
        if not len(urls):
            return []
        # Set a dataframe for tracking the url attempts
        self.tracker = {
            u: {"scraped": False, "attempts": 0}
            for u in urls
        }
        resps = {}
        scrape_urls = set(urls)
        logging.info(f"{len(urls)} unique urls from {len(scrape_urls)}")
        # Start the loop
        self.loop = self._get_event_loop()
        while len(scrape_urls):
            init_len = len(scrape_urls)
            # Ensure shutdown flag is reset self.shutdown_initiated
            self.shutdown_initiated = False
            self.reset_pages_scraped()
            self.total_to_scrape = len(scrape_urls)
            # Gaher tasks and run
            self.coro = self._fetch_all_async(scrape_urls)
            # Build try except clause for premature cancellation
            scrape_resps = self.loop.run_until_complete(self.coro)
            # Process responses
            scrape_urls, resps, failed_urls = \
                self.handle_responses(scrape_urls, resps,
                                      scrape_resps, init_len)
            # Sleep before running again
            # - shutdown must have been initiated
            # - there must still be urls to scrape
            # - the rest between attempt flag must be set to True
            if self.shutdown_initiated \
                    and len(scrape_urls) \
                    and self.rest_between_attempts:
                logging.info(f"Sleeping for {self.rest_wait} seconds")
                sleep(self.rest_wait)
        logging.info(
            f"Scraping complete {len(failed_urls)}/{len(set(urls))} urls failed")
        # Convert resps back
        resps = [v for _, v in resps.items()]
        # end the job
        self.end_job()
        return resps



from time import sleep
from requests_html import HTMLSession
import logging
from requests import Response
from http.client import HTTPResponse

from .base_scrape import BaseScrape
from ..utils.header_vars import random_header_vars
from ..utils.errors import HttpResponseStatusError


class Scrape(BaseScrape):
    def __init__(self,
                 post_process_func: callable,
                 post_process_kwargs: dict = {},
                 fetch_error_handler: callable = None,
                 use_proxy: bool = False,
                 proxy: str = None,
                 pac_url: str = None,
                 consecutive_error_limit: int = 100,
                 attempt_limit: int = 5,
                 rest_between_attempts: bool = True,
                 rest_wait: int = 60,
                 randomise_headers: bool = False
                 ):
        """Class for scrapping webpages

        args:
        ----
        post_process_func - callable - for processing html
        post_process_kwargs - dict:{} - kwargs for use in post processing
        fetch_error_handler - callable:None - the function to be called if an
            error is experienced during _fetch. Passes in:
            url, error as arguments
        use_proxy - bool:False - should a proxy be used
        proxy - str:None - what is the address of the proxy ONLY VALID IF
            PROXY IS TRUE
        pac_Url - str:None - the location of the pac information ONLY VALID IF
            PROXY IS TRUE
        consecutive_error_limit - int:100 - the number of times an error can be experienced 
            in a row before the scrape is cancelled and a new round is started
        attempt_limit - int:5 - number of times a url can be attempted before it's abandoned
        rest_between_attempts - bool:True - should the program rest between scrapes
        rest_wait - int:60 - how long should the program rest for ONLY VALID IF
            REST_BETWEEN_SCRAPES IS TRUE
        randomise_headers - bool:False - should the headers be randomised after each request
        """
        # Init super
        super().__init__(
            use_proxy=use_proxy,
            proxy=proxy,
            pac_url=pac_url
        )
        self.post_process = post_process_func
        self.post_process_kwargs = post_process_kwargs
        self.randomise_headers = randomise_headers
        self.headers = random_header_vars(self.header_vars)
        self.fetch_error_handler = fetch_error_handler
        self.session = HTMLSession()
        # Define allowed errors
        self.acceptable_errors = (HttpResponseStatusError,)
        self.consecutive_error_limit = consecutive_error_limit
        self.consecutive_error_count = 0
        # Define criteria for looping multiple attempts
        self.attempt_limit = attempt_limit
        self.rest_between_attempts = rest_between_attempts
        self.rest_wait = rest_wait
        self.tracker = None
        self.cur_err = None

    def _proxy(self):
        # Start the pac session
        self._get_pac_session()
        self.session = self.pac_session

    def _request(self, url: str):
        return self.session.get(url, headers=self.headers)

    def _fetch(self, url):
        """Function to fetch HTML from url

        args:
        ----
        url - str - url to be requested

        returns:
        ----
        list
        """
        resp = None
        status = None
        # Make the request
        try:
            if url:
                resp = self._request(url)
                if isinstance(resp, Response):
                    status = resp.status_code
                    html = resp.text
                elif isinstance(resp, HTTPResponse):
                    status = resp.status
                    html = resp.read()
                else:
                    raise TypeError(
                        "resp should be of type HTTPResponse or Response")
                if status != 200:
                    raise HttpResponseStatusError(
                        f"url responded with a http status of {status}")
                func_resp = self.post_process(
                    html=html, resp=resp, **self.post_process_kwargs) \
                    if html is not None else None
                # Reset self.acceptable_error_count if all goes fine
                self.consecutive_error_count = 0
                return {"url": url, "func_resp": func_resp, "status": status, "error": None}
        except Exception as e:
            # Set the current error - increment if the same error
            if type(e) == self.cur_err:
                self.consecutive_error_count += 1
            else:
                self.cur_err = type(e)
                self.consecutive_error_count = 1
            # Check if acceptabe error limit has been reached
            # this prevents functions from carrying on after a site has started blocking calls
            if self.consecutive_error_count >= self.consecutive_error_limit:
                logging.warning(
                    f"Consecutive error limit reached - {e} - consecutive count at {self.consecutive_error_count}/{self.consecutive_error_limit}")
            # Check for error handler
            if self.fetch_error_handler:
                logging.info(
                    f"Error passed to {self.fetch_error_handler.__name__}")
                # Run the error handler
                self.fetch_error_handler(url, e)
            # Check if acceptable error
            if type(e) in self.acceptable_errors:
                logging.error(
                    f"Acceptable error in request or post processing {url} - {e}")
            # Raise error
            else:
                logging.error(
                    f"Unhandled error in request or post processing {url} - {e}")
                if f"{e}" == "":
                    raise e
            return {"url": url, "func_resp": None, "status": status, "error": e}

    # run from terminal
    def scrape_all(self, urls: list):
        """"Function scraping html from urls and passing 
        them through the post processing function

        args:
        ----
        urls - list - the pages to be scraped

        returns:
        ----
        list of dicts
            EG [{
                "url":"http://google.com",
                "success":True,
                "status":200
            }]
        """
        self.start_job()
        if self.use_proxy:
            self._proxy()
        # Establish urls
        if not len(urls):
            return []
        # Set a dataframe for tracking the url attempts
        self.tracker = {
            u: {"scraped": False, "attempts": 0}
            for u in urls
        }
        resps = {}
        scrape_urls = set(urls)
        logging.info(f"{len(urls)} unique urls from {len(scrape_urls)}")
        while len(scrape_urls):
            init_len = len(scrape_urls)
            self.reset_pages_scraped()
            self.total_to_scrape = len(scrape_urls)
            # Run the scrapes
            scrape_resps = []
            for url in scrape_urls:
                scrape_resps.append(self._fetch(url))
                self.increment_pages_scraped()
                # Regenerate headers
                if self.randomise_headers:
                    self.headers = random_header_vars(self.header_vars)
            # Process responses
            scrape_urls, resps, failed_urls = \
                self.handle_responses(scrape_urls, resps,
                                      scrape_resps, init_len)
            # Sleep before running again
            # - shutdown must have been initiated
            # - there must still be urls to scrape
            # - the rest between attempt flag must be set to True
            if len(scrape_urls) \
                    and self.rest_between_attempts:
                logging.info(f"Sleeping for {self.rest_wait} seconds")
                sleep(self.rest_wait)
        logging.info(
            f"Scraping complete {len(failed_urls)}/{len(set(urls))} urls failed")
        # Convert resps back
        resps = [v for _, v in resps.items()]
        # end the job
        self.end_job()
        return resps

    # run from terminal
    def scrape_one(self, url: str):
        """"Function scraping html from a single url and passing 
        it through the post processing function

        args:
        ----
        url - str - the pages to be scraped

        returns:
        ----
        list of dicts
            EG [{
                "url":"http://google.com",
                "success":True,
                "status":200
            }]
        """
        if self.use_proxy:
            self._proxy()
        for _ in range(self.attempt_limit):
            resp = self._fetch(url)
            if resp["status"]:
                break
            elif self.rest_between_attempts:
                logging.info(f"Sleeping for {self.rest_wait} seconds")
                sleep(self.rest_wait)
        return resp



"""All methods relating to beautiful soup"""

def tree_find_all(soup, find_params:list):
    """Will descend tree collecting all results
    
    params are in format:
    [tag, {attrs}]
    """
    soups = [soup]
    for params in find_params:
        out = []
        for s in soups: 
            #Find all in current soup
            mini_soup = s.find_all(params[0], attrs=params[1])
            out.extend(mini_soup)
        soups = out.copy()
    return soups


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['async_scrape', 'async_scrape.libs', 'async_scrape.utils']

package_data = \
{'': ['*']}

install_requires = \
['Brotli>=1.0.9,<2.0.0',
 'PyPAC>=0.16.0,<0.17.0',
 'aiohttp>=3.8.3,<4.0.0',
 'nest-asyncio>=1.5.6,<2.0.0',
 'requests-html>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'async-scrape',
    'version': '0.1.14',
    'description': 'A package designed to scrape webpages using aiohttp and asyncio. Has some error handling to overcome common issues such as sites blocking you after n requests over a short period.',
    'long_description': '# Async-scrape\n## _Perform webscrape asyncronously_\n\n[![Build Status](https://travis-ci.org/joemccann/dillinger.svg?branch=master)](https://travis-ci.org/joemccann/dillinger)\n\nAsync-scrape is a package which uses asyncio and aiohttp to scrape websites and has useful features built in.\n\n## Features\n\n- Breaks - pause scraping when a website blocks your requests consistently\n- Rate limit - slow down scraping to prevent being blocked\n\n\n## Installation\n\nAsync-scrape requires [C++ Build tools](https://go.microsoft.com/fwlink/?LinkId=691126) v15+ to run.\n\n\n```\npip install async-scrape\n```\n\n## How to use it\n```\n#Create an instance\nfrom async_scrape import AsyncScrape\n\ndef post_process(html, resp, **kwargs):\n    """Function to process the gathered response from the request"""\n    if resp.status == 200:\n        return "Request worked"\n    else:\n        return "Request failed"\n\nasync_Scrape = AsyncScrape(\n    post_process_func=post_process,\n    post_process_kwargs={},\n    fetch_error_handler=None,\n    use_proxy=False,\n    proxy=None,\n    pac_url=None,\n    acceptable_error_limit=100,\n    attempt_limit=5,\n    rest_between_attempts=True,\n    rest_wait=60\n)\n\nurls = [\n    "https://www.google.com",\n    "https://www.bing.com",\n]\n\nresps = async_Scrape.scrape_all(urls)\n```\n\nResponse object is a list of dicts in the format:\n```\n{\n    "url":url, #url of request\n    "func_resp":func_resp, #response from post processing function\n    "status":resp.status, #http status\n    "error":None #any error encountered\n}\n```\n\n\n## License\n\nMIT\n\n**Free Software, Hell Yeah!**\n\n[//]: # (These are reference links used in the body of this note and get stripped out when the markdown processor does its job. There is no need to format nicely because it shouldn\'t be seen. Thanks SO - http://stackoverflow.com/questions/4823468/store-comments-in-markdown-syntax)\n\n   [dill]: <https://github.com/joemccann/dillinger>\n   [git-repo-url]: <https://github.com/joemccann/dillinger.git>\n   [john gruber]: <http://daringfireball.net>\n   [df1]: <http://daringfireball.net/projects/markdown/>\n   [markdown-it]: <https://github.com/markdown-it/markdown-it>\n   [Ace Editor]: <http://ace.ajax.org>\n   [node.js]: <http://nodejs.org>\n   [Twitter Bootstrap]: <http://twitter.github.com/bootstrap/>\n   [jQuery]: <http://jquery.com>\n   [@tjholowaychuk]: <http://twitter.com/tjholowaychuk>\n   [express]: <http://expressjs.com>\n   [AngularJS]: <http://angularjs.org>\n   [Gulp]: <http://gulpjs.com>\n\n   [PlDb]: <https://github.com/joemccann/dillinger/tree/master/plugins/dropbox/README.md>\n   [PlGh]: <https://github.com/joemccann/dillinger/tree/master/plugins/github/README.md>\n   [PlGd]: <https://github.com/joemccann/dillinger/tree/master/plugins/googledrive/README.md>\n   [PlOd]: <https://github.com/joemccann/dillinger/tree/master/plugins/onedrive/README.md>\n   [PlMe]: <https://github.com/joemccann/dillinger/tree/master/plugins/medium/README.md>\n   [PlGa]: <https://github.com/RahulHP/dillinger/blob/master/plugins/googleanalytics/README.md>\n',
    'author': 'Robert Franklin',
    'author_email': 'cia05rf@gmail.com',
    'maintainer': 'Robert Franklin',
    'maintainer_email': 'cia05rf@gmail.com',
    'url': 'https://github.com/cia05rf/async-scrape/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime, timedelta
from pypac import PACSession, get_pac
import sys
import re
import logging
from time import sleep


class BaseScrape:
    def __init__(self,
                 use_proxy: bool = False,
                 proxy: str = None,
                 pac_url: str = None
                 ):
        """Class for scrapping webpages

        args:
        ----
        use_proxy - bool:False - should a proxy be used
        proxy - str:None - what is the address of the proxy ONLY VALID IF
            PROXY IS TRUE
        pac_Url - str:None - the location of the pac information ONLY VALID IF
            PROXY IS TRUE
        """
        self.pages_scraped = 0
        self.total_to_scrape = 0
        self.job_start = None
        self.job_end = None
        self.time_marks = []
        self.use_proxy = use_proxy
        self.proxy = proxy
        self.pac = get_pac(url=pac_url) \
            if self.use_proxy and pac_url is not None else None
        self.pac_session = None
        # Variables for randomly generating headers
        self.header_vars = None

    def _deconstruct_url(self, url):
        if re.search(r"^http://", url):
            return "http", url[7:]
        elif re.search(r"^https://", url):
            return "https", url[8:]
        else:
            raise ValueError(f"Invalid url -> {url}")

    def _get_pac_session(self):
        if not self.pac_session:
            self.pac_session = PACSession(self.pac)
        return self.pac_session

    def _get_proxies(self, url):
        if self.use_proxy:
            # use pypac
            proxies = self.pac_session \
                ._get_proxy_resolver(self.pac) \
                .get_proxy_for_requests(url)
        else:
            proxies = None
        return proxies

    def _get_proxy(self, url):
        if self.use_proxy:
            if self.proxy:
                # use given proxy
                proxy = self.proxy
            elif self.pac:
                # use pypac
                proxies = self._get_proxies(url)
                match = re.search(r"^(\w*)", str(url))
                proxy = proxies[match.group()]
            else:
                raise ValueError(
                    "Either pac_url or a proxy must being given in order for use_proxy to be True")
        else:
            proxy = None
        return proxy

    def handle_responses(self, scrape_urls, resps, scrape_resps, init_len):
        # Add scrape_resps to resps
        resps |= {r["url"]: r for r in scrape_resps}
        # Get scraped urls
        # Split success and fails
        success_urls = set([
            r["url"] for r in scrape_resps
            if not r["error"]
        ])
        errored_urls = set([
            r["url"] for r in scrape_resps
            if r["error"]
        ])
        # Increment attempts count on each scraped url
        self._increment_attempts(True, scrape_urls)
        # Increment attempts count on each attempted but failed (IE had an error
        # but not cancelled)
        self._increment_attempts(False, errored_urls)
        # Remove scraped urls from scrape_urls
        scrape_urls = scrape_urls.difference(success_urls)
        # Remove urls where too many attempts have been made
        failed_urls = set(k for k, v in self.tracker.items()
        if v["attempts"] >= self.attempt_limit)
        scrape_urls = scrape_urls.difference(failed_urls)
        logging.info(f"""Scraping round complete, summary:
    attempted:                               {init_len}
    successful scrapes:                      {len(success_urls)}
    errored scrapes (will attempt again):    {len(errored_urls)}
    failed scrapes (will not attempt again): {len(failed_urls)}
    remaining urls:                          {len(scrape_urls)}""")
        return [scrape_urls, resps, failed_urls]

    def _increment_attempts(self, scraped: bool, urls: list = []):
        for u in urls:
            self.tracker[u]["scraped"] = scraped
            self.tracker[u]["attempts"] += 1

    def reset_pages_scraped(self):
        self.pages_scraped = 0
        self.time_marks = [datetime.now()]

    def increment_pages_scraped(self):
        self.pages_scraped += 1
        # Add time mark
        self.time_marks.append(datetime.now())
        # Calc loop time
        loop_time_elapsed = (
            self.time_marks[-1] - self.time_marks[-2]).total_seconds()
        # Calc estimated finish time
        total_time_elapsed = (datetime.now() - self.job_start).total_seconds()
        est_total_time_s = total_time_elapsed + \
            (self.total_to_scrape * total_time_elapsed / self.pages_scraped)
        # Output
        sys.stdout.write(
            f"\rprocessed -> {self.pages_scraped}/{self.total_to_scrape} - loop time {loop_time_elapsed:.6f} - est finish {datetime.now() + timedelta(seconds=est_total_time_s)}\n")
        sys.stdout.flush()

    def throttle_tasks(self):
        """Introduces sleep between tasks to slow them to a throttled rate"""
        pass

    def start_job(self):
        self.job_start = datetime.now()
        self.time_marks = [self.job_start]
        self.job_end = None

    def end_job(self):
        self.job_end = datetime.now()
        runtime = self.job_end - self.job_start
        logging.info(f"Job completed in {runtime}")

    def _get_pac_session(self):
        if not self.pac_session:
            self.pac_session = PACSession(self.pac)
        return self.pac_session



"""List of vars to use in headers"""
import random
from secrets import randbelow
from typing import Dict
import re

HEADER_VARS = {
    "Accept": ["text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"],
    "Accept-Encoding": ["gzip, deflate, br"],
    "Accept-Language": ["en-GB,en-US;q=0.9,en;q=0.8"],
    "Dnt": ["1"],
    "Upgrade-Insecure-Requests": ["1"],
    "User-Agent": [
        "Mozilla/5.0 (Linux; Android 12; SM-S906N Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 10; SM-G996U Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 10; SM-G980F Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.96 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 9; SM-G973U Build/PPR1.180610.011) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 8.0.0; SM-G960F Build/R16NW) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.84 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.0; SM-G892A Build/NRD90M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/60.0.3112.107 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.0; SM-G930VC Build/NRD90M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/58.0.3029.83 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0.1; SM-G935S Build/MMB29K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/55.0.2883.91 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0.1; SM-G920V Build/MMB29K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.98 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 5.1.1; SM-G928X Build/LMY47X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.83 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 12; Pixel 6 Build/SD1A.210817.023; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/94.0.4606.71 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 11; Pixel 5 Build/RQ3A.210805.001.A1; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/92.0.4515.159 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 10; Google Pixel 4 Build/QD1A.190821.014.C2; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.108 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 10; Google Pixel 4 Build/QD1A.190821.014.C2; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.108 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2 Build/OPD1.170811.002; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/59.0.3071.125 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.1.1; Google Pixel Build/NMF26F; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/54.0.2840.85 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 6P Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.83 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 9; J8110 Build/55.0.A.0.552; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/71.0.3578.99 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.1.1; G8231 Build/41.2.A.0.219; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/59.0.3071.125 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0.1; E6653 Build/32.2.A.0.253) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.98 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 10; HTC Desire 21 pro 5G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.127 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 10; Wildfire U20 5G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.136 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0; HTC One X10 Build/MRA58K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/61.0.3163.98 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0; HTC One M9 Build/MRA58K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.98 Mobile Safari/537.3",
        "Mozilla/5.0 (iPhone14,6; U; CPU iPhone OS 15_4 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/19E241 Safari/602.1",
        "Mozilla/5.0 (iPhone14,3; U; CPU iPhone OS 15_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/19A346 Safari/602.1",
        "Mozilla/5.0 (iPhone13,2; U; CPU iPhone OS 14_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/15E148 Safari/602.1",
        "Mozilla/5.0 (iPhone12,1; U; CPU iPhone OS 13_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/15E148 Safari/602.1",
        "Mozilla/5.0 (iPhone12,1; U; CPU iPhone OS 13_0 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/15E148 Safari/602.1",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1", "Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/69.0.3497.105 Mobile/15E148 Safari/605.1",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/13.2b11866 Mobile/16A366 Safari/605.1.15",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A5370a Safari/604.1",
        "Mozilla/5.0 (iPhone9,3; U; CPU iPhone OS 10_0_1 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14A403 Safari/602.1",
        "Mozilla/5.0 (iPhone9,4; U; CPU iPhone OS 10_0_1 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14A403 Safari/602.1",
        "Mozilla/5.0 (Apple-iPhone7C2/1202.466; U; CPU like Mac OS X; en) AppleWebKit/420+ (KHTML, like Gecko) Version/3.0 Mobile/1A543 Safari/419.3",
        "Mozilla/5.0 (Windows Phone 10.0; Android 6.0.1; Microsoft; RM-1152) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Mobile Safari/537.36 Edge/15.15254",
        "Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; RM-1127_16056) AppleWebKit/537.36(KHTML, like Gecko) Chrome/42.0.2311.135 Mobile Safari/537.36 Edge/12.10536",
        "Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Microsoft; Lumia 950) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/13.1058",
        "Mozilla/5.0 (Linux; Android 12; SM-X906C Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119 Mobile Safari/537.36",
        "Mozilla/5.0 (Linux; Android 11; Lenovo YT-J706X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.0; Pixel C Build/NRD90M; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/52.0.2743.98 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0.1; SGP771 Build/32.2.A.0.253; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/52.0.2743.98 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 6.0.1; SHIELD Tablet K1 Build/MRA58K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/55.0.2883.91 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.0; SM-T827R4 Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.116 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 5.0.2; SAMSUNG SM-T550 Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/3.3 Chrome/38.0.2125.102 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 4.4.3; KFTHWI Build/KTU84M) AppleWebKit/537.36 (KHTML, like Gecko) Silk/47.1.79 like Chrome/47.0.2526.80 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 5.0.2; LG-V410/V41020c Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/34.0.1847.118 Safari/537.36",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.246",
        "Mozilla/5.0 (X11; CrOS x86_64 8172.45.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.64 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_2) AppleWebKit/601.3.9 (KHTML, like Gecko) Version/9.0.2 Safari/601.3.9",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.111 Safari/537.36",
        "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:15.0) Gecko/20100101 Firefox/15.0.1",
        "Mozilla/5.0 (PlayStation; PlayStation 5/2.26) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0 Safari/605.1.15",
        "Mozilla/5.0 (PlayStation 4 3.11) AppleWebKit/537.73 (KHTML, like Gecko)",
        "Mozilla/5.0 (PlayStation Vita 3.61) AppleWebKit/537.73 (KHTML, like Gecko) Silk/3.2",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Xbox; Xbox Series X) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/48.0.2564.82 Safari/537.36 Edge/20.02",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; XBOX_ONE_ED) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.79 Safari/537.36 Edge/14.14393",
        "Mozilla/5.0 (Windows Phone 10.0; Android 4.2.1; Xbox; Xbox One) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/13.10586",
        "Mozilla/5.0 (Nintendo Switch; WifiWebAuthApplet) AppleWebKit/601.6 (KHTML, like Gecko) NF/4.0.0.5.10 NintendoBrowser/5.1.0.13343",
        "Mozilla/5.0 (Nintendo WiiU) AppleWebKit/536.30 (KHTML, like Gecko) NX/3.0.4.2.12 NintendoBrowser/4.3.1.11264.US",
        "Mozilla/5.0 (Nintendo 3DS; U; ; en) Version/1.7412.EU",
        "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
        "Mozilla/5.0 (compatible; bingbot/2.0; +http://www.bing.com/bingbot.htm)",
        "Mozilla/5.0 (compatible; Yahoo! Slurp; http://help.yahoo.com/help/us/ysearch/slurp)",
        "Mozilla/5.0 (X11; U; Linux armv7l like Android; en-us) AppleWebKit/531.2+ (KHTML, like Gecko) Version/5.0 Safari/533.2+ Kindle/3.0+",
        "Mozilla/5.0 (Linux; U; en-US) AppleWebKit/528.5+ (KHTML, like Gecko, Safari/528.5+) Version/4.0 Kindle/3.0 (screen 600x800; rotate)"
    ]
}

def gen_client_hint(ua:str) -> str:
    """Generates the client hint from user-agent string"""
    # "sec-ch-ua": '"Google Chrome";v="98", "Not)A;Brand";v="8", "Chromium";v="98"'
    # "sec-ch-ua-mobile": '?0'
    # 'sec-ch-ua-platform': '"Windows"',
    # Get platform and mobile
    ch_ua = {}
    os_group = re.search("(?<=^Mozilla/5.0 \()([^\)]+)(?=\))", ua)
    if os_group is not None:
        os_group = os_group.group()
        os_group = os_group.split(";")
        if len(os_group):
            platform = os_group[0].strip()
            if len(os_group) > 1:
                os = os_group[1]
                if "iPhone" in os or "iPhone" in platform \
                    or "Android" in os \
                    or "Windows Phone" in os:
                    mobile = "?1"
                else:
                    mobile = "?0"
            else:
                mobile = "?0"
    else:
        platform = "Linux"
        mobile = "?0"
    ch_ua["Sec-CH-UA-Mobile"] = mobile
    ch_ua["Sec-CH-UA-Platform"] = f'"{platform}"'

    # Get the user agent
    BROWSERS = {
        "Chrome": "Google Chrome",
        "Edge": "Microsoft Edge",
        "Opera": "Opera",
        "Safari": "Safari",
        "Firefox": "Firefox"
    }
    CHROMIUM = ["Google Chrome", "Microsoft Edge", "Opera"]
    # Collect all the agents
    ua_group = re.search("(?<=\(KHTML, like Gecko\)).+", ua)
    def create_ua_str(s:str):
        split_s = s.split("/")
        try:
            # Extract the browser
            if split_s[0] in list(BROWSERS.keys()):
                browser_name = BROWSERS[split_s[0].strip()]
                is_chromium = True if browser_name in CHROMIUM else False
            else:
                browser_name = ' Not A;Brand'
                is_chromium = False
            # Extract the version
            version_match = re.search("^(\d+)[^\d]", split_s[1])
            version = int(version_match.group(1)) if version_match is not None else 1
            chromium = [f'"Chromium";v="{version}"'] if is_chromium else []
            browser = f'"{browser_name}";v="{version}"'
            return ", ".join([browser] + chromium)
        except:
            return None

    if ua_group is not None:
        ua_group = ua_group.group()
        ua_group = ua_group.split(" ")
        us_group = [create_ua_str(s) for s in ua_group]
        ua_s = ", ".join([s for s in us_group if s is not None]).strip()
        ch_ua["Sec-CH-UA"] = ua_s
    return ch_ua


def random_header_vars(header_vars: Dict[str, list] = None, seed: int = None):
    """Gets a set of random headers to use"""
    if seed is not None:
        random.seed(seed)
    headers = {}
    if header_vars is None:
        header_vars = HEADER_VARS
    for k, v in header_vars.items():
        if len(v) == 1:
            headers[k] = v[0]
            continue
        i = random.randrange(len(v))
        headers[k] = v[i]
    # Generate client hints
    if "User-Agent" in headers:
        ch_ua = gen_client_hint(headers["User-Agent"])
        headers |= ch_ua
    return headers
